

#include "stdio.h"

#include "ad++.h"
#include "admacros.h"
//#include "adftn.h"

#include <time.h>


//!-----------------------------------------------------------------------
//!
//!  Q State 2d Potts Model in C++ with parallel common runtime
//!  system Adlib support.
//!
//!
//!-----------------------------------------------------------------------
   
      
void main(int argc, char **argv) {
        void setup_random( unsigned short * xsubi , int iseed);

  AdlibInit(argc, argv) ;
  int nprocs = AdlibNProcs() ;

  Procs1 p(nprocs) ;

  ON(p) {
    int m = 64;
    int n = 64;

    Range x(m) ;
    Range y(n, p.dim(0), BLK) ;


    int q = 4;

    float cup = -4.0;

    Array2<int> latt(x, y) ;
    Array2<int> sh_tmp(x, y) , sh_tmp1(x,y),sh_tmp2(x,y),sh_tmp3(x,y);
    Array2<int> sh_tmp4(x, y) , sh_tmp5(x,y),sh_tmp6(x,y),sh_tmp7(x,y);
    Array2<char> result1(x, y) ;
    Array2<char> result2(x, y) ;
    Array2<char> result3(x, y) ;

    Array2<int> itmp(x, y) ;

    Array2<float> tmp(x, y) ;
    Array2<float> p(x, y) ;

    Array2<int> mask(x, y) ;

    int nmeasure = 2;
    int niter = 10;

    Index i(x), j(y);
    unsigned short xsubi[3];

    float t1, t2;

    setup_random( xsubi, 12345 );

    WHERE(i) {
      WHERE(j) {
        tmp(i, j) = erand48(xsubi);
        latt(i, j) = int( tmp(i, j) * q);
      } EREWH(j);
    } EREWH(i);

    for (int imeasure = 1; imeasure<=nmeasure; imeasure++) {

      // Measurements:

      shift( sh_tmp, latt, 1, 0, CYCL);
      WHERE(i) {
        WHERE(j) {
          itmp(i, j) = latt(i, j) == sh_tmp(i, j) ? 1 : 0 ;
        } EREWH(j);
      } EREWH(i);

      shift(sh_tmp, latt, -1, 0, CYCL);
      WHERE(i) {
        WHERE(j) {
        if (latt(i, j) == sh_tmp(i, j)) itmp(i, j) += 1;
        } EREWH(j);
      } EREWH(i);

      shift(sh_tmp, latt, 1, 1, CYCL);
      WHERE(i) {
        WHERE(j) {
          if (latt(i, j) == sh_tmp(i, j)) itmp(i, j) += 1;
        } EREWH(j);
      } EREWH(i);

      shift(sh_tmp, latt, -1, 1, CYCL);
      WHERE(i) {
        WHERE(j) {
          if (latt(i, j) == sh_tmp(i, j)) itmp(i, j) += 1;
        } EREWH(j);
      } EREWH(i);

      gprintf("\t# Bonds = %d\n",sum( itmp ));

      t1 = (float)clock()/CLOCKS_PER_SEC;

      for (int iter=1; iter<=niter; iter++) {

        WHERE(i) {
          WHERE(j) {
            tmp(i, j) = erand48(xsubi);
            itmp(i, j) = (latt(i, j) + int(tmp(i, j) * q-1) + 1) % q ;
          } EREWH(j);
        } EREWH(i);


        for (int i0=1; i0<=2; i0++) {
          for (int j0=1; j0<=2; j0++) {


	    WHERE(i) {
	      WHERE(j) {
		tmp(i, j) = erand48(xsubi);
		mask(i, j) = 0;
	      } EREWH(j);
	    } EREWH(i);


	    Index i2(x.subrng((m - i0) / 2 + 1, i0 - 1 , 2)) ;
	    Index j2(y.subrng((m - j0) / 2 + 1, j0 - 1 , 2)) ;
	    WHERE(i2) {
	      WHERE(j2) {
	        mask(i2, j2) = 1 ;
	      } EREWH(j2);
	    } EREWH(i2);

            shift(sh_tmp, latt, 1, 0, CYCL);
            shift(sh_tmp1, latt, -1, 0, CYCL);
            shift(sh_tmp2, latt, 1, 1, CYCL);
	    shift(sh_tmp3, latt, -1, 1, CYCL);
	    shift(sh_tmp4, itmp, 1, 0, CYCL);
	    shift(sh_tmp5, itmp, -1, 0, CYCL);
	    shift(sh_tmp6, itmp, 1, 1, CYCL);
	    shift(sh_tmp7, itmp, -1, 1, CYCL);
            WHERE(i) {
              WHERE(j) {
                if (mask(i, j)) {
	  	  p(i, j) = latt(i, j) == sh_tmp(i, j) ? cup : 0.0 ;
                  if (latt(i, j) == sh_tmp1(i, j)) p(i, j) += cup;
                  if (latt(i, j) == sh_tmp2(i, j)) p(i, j) += cup;
		  if (latt(i, j) == sh_tmp3(i, j)) p(i, j) += cup;
		  if (latt(i, j) == sh_tmp4(i, j)) p(i, j) -= cup;
		  if (latt(i, j) == sh_tmp5(i, j)) p(i, j) -= cup;
		  if (latt(i, j) == sh_tmp6(i, j)) p(i, j) -= cup;
		  if (latt(i, j) == sh_tmp7(i, j)) p(i, j) -= cup;
                  p(i, j) = exp( p(i, j) );
                  if (tmp(i, j) < p(i, j)) latt(i, j) = itmp(i, j);
                }
	      } EREWH(j);
	    } EREWH(i);
          }
        }
      }
      t2 = (float)clock()/CLOCKS_PER_SEC;
      gprintf("Time: %f\n", t2 - t1);
    }       

    FILE* ifile = gfopen("out.ppm", "w") ;

    gfprintf(ifile, "P6\n");
    gfprintf(ifile, "# 2d Potts Model\n");
    gfprintf(ifile, "%12d%12d\n",m,n);
    gfprintf(ifile, "%8d\n",127);

    WHERE(i) {
      WHERE(j) {
        switch (latt(i, j)) {
          case 0 :
            result1(i, j) = 127;
            result2(i, j) = 0;
            result3(i, j) = 0;
            break;
          case 1 :
            result1(i, j) = 0;
            result2(i, j) = 127;
            result3(i, j) = 0;
            break;
          case 2 :
            result1(i, j) = 0;
            result2(i, j) = 0;
            result3(i, j) = 127;
            break;
          case 3 :
            result1(i, j) = 127;
            result2(i, j) = 127;
            result3(i, j) = 0;
            break;
          case 4 :
            result1(i, j) = 127;
            result2(i, j) = 0;
            result3(i, j) = 127;
            break;
          case 5 :
            result1(i, j) = 0;
            result2(i, j) = 127;
            result3(i, j) = 127;
            break;
          case 6 :
            result1(i, j) = 127;
            result2(i, j) = 127;
            result3(i, j) = 127;
            break;
          default :
            result1(i, j) = 0;
            result2(i, j) = 0;
            result3(i, j) = 0;
            break;
        }
      } EREWH(j);
    } EREWH(i);
        
    afprintf(ifile, "%c%c%c",(result1),(result2),(result3));
        
    gfclose(ifile);

  } NO(p);

  AdlibFinalize() ;
}


/*
        
      end program 

!-----------------------------------------------------------------------

      subroutine setup_random(iseed)  ! put a new seed in the bloody awful system supplied RNG
      integer :: n
      integer, dimension(:), allocatable ::  seed
      call random_seed(size=n)
      allocate( seed(n) )
      forall(i=1:n) seed(i) = mod(i*4096+150889+iseed,714025 )
      call random_seed( put=seed(1:n) )
      deallocate( seed )
      end subroutine

*/

void setup_random( unsigned short * xsubi , int iseed)
{
        int procs = AdlibLocId();
        
        for (int i=0; i<3; i++) {
                xsubi[i] = (procs * 4096 + 889 + iseed) % 24025;
                procs *= 3;
        }
}

